/*
DL2 Monitor: ./mon/dl2mon.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (dl2mon): DL2 Monitor Process
 *
 * The process reads data from a packet socket and writes them in raw format
 * to an outfile. It is called the DL2 Monitor and has been designed to operate
 * in connection with a peer process, referred to as DL2 Analyzer. There are
 * two modes of operation: online and offline. In online mode, the Monitor is
 * connected to the peer by means of a pipe, with the Analyzer being the child. 
 * In offline mode, the Monitor doesn't know of any peer and only writes to a 
 * regular file what has coming in from the packet socket. The online operation
 * is considered the default. In online mode, the peers are also connected via 
 * shared memory in order to transfer some status information. The Monitor acts
 * as the creator of that memory.
 *
 * Online Mode
 * Assume that all command line and rcfile arguments have been parsed without
 * errors. Then, the process forks a child which executes the DL2 Analyzer, 
 * connects to it uni-directional via pipe mechanism, attaches to memory shared
 * with that child, writes some arguments to that memory and waits for the peer
 * has coming up. Then, it opens the socket. Data will be transferred until the
 * parent process is killed by SIGTERM or SIGINT.  Messages always will appear 
 * on stderr.
 *
 * Offline Mode
 * This mode must be specified explicitly in conjunction with a regular output
 * file which will receive data from the socket. There is no difference to the
 * online mode, except that no peers will be invoked. Hence, there is no need
 * to use IPC. 
 *
 * The default socket-type ist DL2_ST_RAW, the default protocol is DL2_PD_ALL.
 * Known socket-types: according to DL2_KEY_SOCKTYPE_* settings;
 * known protocols: according to DL2_KEY_PROTOCOL_* settings.
 * The default outfile is read from the .dl2monrc file and considered to
 * exist for write-only operations, i.e. the monitor doesn't create it. The 
 * outfile specified by command line arguments will be created if it doesn't 
 * exist. Otherwise, the monitor goes to terminate.
 *
 * Usage: dl2mon {monitor-option} [-A {analyzer-option}]
 *       local monitor-options:
 *   -s  socket-type
 *   -p  protocol-name 
 *   -P  protocol-number 
 *   -i  interface-name 
 *   -o  outfile (offline)
 *       others from lib/monpa.h: 
 *   -C  rcfile 
 *   -F{<flag>}
 *    o  force offline mode
 *
 * Implementation Notes:
 * In online mode, the parent writes to stdout, the child reads from stdin.
 * The SIGCHLD is catched in order to terminate on loss of child, for some
 * reasons. Normally, parent sends here-we-are and waits for response from
 * the child. Then, transmission begins and will continue until SIGTERMed.
 *  
 * -cpid: process id of the DL2 Analyzer (online);
 * -monfd, outfile: descriptor of the Monitor output file;
 * -sockfd: descriptor of the packet socket;
 * -shmid, shmmonpa: id and data of the shmy (online);
 * -terminate: flag indicates to terminate, set by SIGTERM, SIGINT signal;
 * -socket_type, protocol: arguments for the packet socket;
 * -dl2monrc: filename of the rcfile
 * -options: status information from arguments parsing;
 *
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020504  execve replaced by execvp: searching for dl2pa is left to exec.
 * 20020503  dl2pa: bug fixed with name generation.
 * 20020501  DL2_VERSION: package version introduced.
 * 20020411  online: -A analyzer options introduced.
 * 20020411  online: connection to analyzer via fork and pipe.
 * 20020410  shared memory usage introduced.
 * 20020408  Dl2_Monheader filled in, here.  
 * 20020408  network interface introduced.  
 * 20020407  command line arguments, rcfile reading introduced.  
 * 20020330  created from scratch
 * $Id: dl2mon.c,v 1.11 2002/05/04 17:36:01 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/ipc.h>
# include <sys/shm.h>
# include <fcntl.h>
# include <signal.h>
# include <errno.h>
# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include <stdarg.h>
# include "lib/dl2mon.h"
# include "lib/monpa.h"

# define COMMENT_MARK DL2_COMMENT_MARK
# define PACKETT_MARK DL2_PACKETT_MARK
# define PARAMST_MARK DL2_PARAMST_MARK
# define RCFILE DL2_RCFILE
# define ENVIRON DL2_ENVIRON
# define NSTR ""

# define OPT_USAGEM   DL2_OPT_USAGEM
# define OPT_RCFILE   DL2_OPT_RCFILE
# define OPT_FLAGS    DL2_OPT_FLAGS
# define OPT_SOCKTYPE "-s"
# define OPT_PROTNAME "-p"
# define OPT_PROTNUM  "-P"
# define OPT_OUTFILE  "-o"
# define OPT_NETIFACE "-i"
# define OPT_ANALYZER "-A"

static int monfd=-1, shmid=-1, sockfd=-1, cpid=-1, terminate=0;
static Dl2_Monpashm *shmmonpa = NULL;
static int socket_type = DL2_ST_RAW, protocol = DL2_PD_ALL;
static int options = 0;
# define SEL_NOPIPE    0x00000001
# define SEL_RCFILE    0x00000002
# define SEL_SOCKTYPE  0x00000004
# define SEL_PROTNAME  0x00000008
# define SEL_OUTFILE   0x00000010
# define SEL_OUTFCREAT 0x00000020
# define SEL_NETIFACE  0x00000040
# define SEL_PROTNUM   0x00000080
# define SEL_OFFLINE   0x00000100
# define SEL_ANALYZER  0x00000200

# define MAXN_ANALYZER 50

static const char* locate_dl2monrc(void);
static const char* read_keyword(const char*,const char*); 
static int attach_shm(const char*);
static int nattach_shm(int);
static int detach_shm(void);
static char* str2literal(char*,char**);
static void message(char*,...);
static void print_usage(void);
static void on_terminate(int); 


int main(int argc,char *argv[]) {
  int rc, i;
  const char *dl2mon = argv[0];
  const char *dl2monrc;
  const char *outfile;
  const char *socket_type_name;
  const char *protocol_name;
  const char *ifacename = NULL;
  const char *analyzer_argv[4+MAXN_ANALYZER];
  const char **analyzer_opt = analyzer_argv+1;
  int n_analyzer_opt = 0;  
  int offline = 0;

  // read command line arguments:
  for (i = 1; i < argc; i++) {
    if (options & SEL_ANALYZER) {
      if (n_analyzer_opt < MAXN_ANALYZER) {
        analyzer_opt[n_analyzer_opt++] = argv[i];
      }
    }
    else if (strcmp(argv[i],OPT_USAGEM) == 0) {
      print_usage(); return 0;
    }
    else if (strcmp(argv[i],OPT_RCFILE) == 0 && i+1 < argc) {
      dl2monrc = argv[++i];
      options |= SEL_RCFILE;
    }
    else if (strcmp(argv[i],OPT_SOCKTYPE) == 0 && i+1 < argc) {
      socket_type_name = argv[++i];
      options |= SEL_SOCKTYPE;
    }
    else if (strcmp(argv[i],OPT_PROTNAME) == 0 && i+1 < argc) {
      protocol_name = argv[++i];
      options &= ~SEL_PROTNUM; protocol = DL2_PD_ALL;
      options |= SEL_PROTNAME;
    }
    else if (strcmp(argv[i],OPT_PROTNUM) == 0 && i+1 < argc) {
      int p;
      if ((sscanf(argv[++i],"%d",&p) == 1) && (p >= 0)) {
        options &= ~SEL_PROTNAME; protocol = p;
        options |= SEL_PROTNUM;
      }
      else {
        message("%s: parse error\n",OPT_PROTNUM);
        return -1;
      }
    }
    else if (strcmp(argv[i],OPT_OUTFILE) == 0 && i+1 < argc) {
      outfile = argv[++i];
      options |= SEL_OUTFILE;
    }
    else if (strcmp(argv[i],OPT_NETIFACE) == 0 && i+1 < argc) {
      ifacename = argv[++i];
      options |= SEL_NETIFACE;
    }
    else if (strncmp(argv[i],OPT_FLAGS,2) == 0) {
      if (strchr(argv[i],'o') != NULL) options |= SEL_OFFLINE;
    }
    else if (strcmp(argv[i],OPT_ANALYZER) == 0 && i+1 < argc) {
      if (n_analyzer_opt < MAXN_ANALYZER) {
        analyzer_opt[n_analyzer_opt++] = argv[++i];
      } options |= SEL_ANALYZER;
    }
    else {
      message("%s: unknown option\n",argv[i]);
      return -1;
    }
  }

  // check command line arguments:
  if ((options & SEL_RCFILE) == 0) {
    dl2monrc = locate_dl2monrc();
  }
  if (dl2monrc != NULL) {
    options |= SEL_RCFILE;
    if ((options & SEL_SOCKTYPE) == 0) {
      socket_type_name = read_keyword(dl2monrc,DL2_KEY_SOCKTYPE);
      if (socket_type_name == NULL) {
        message("error: reading socket-type from %s\n",dl2monrc);
        return -1;
      } options |= SEL_SOCKTYPE;
    }
    if ((options & (SEL_PROTNAME|SEL_PROTNUM)) == 0) {
      protocol_name = read_keyword(dl2monrc,DL2_KEY_PROTOCOL);
      if (protocol_name == NULL) {
        message("error: reading protocol from %s\n",dl2monrc);
        return -1;
      } options |= SEL_PROTNAME;
    }
    if ((options & SEL_OUTFILE) == 0) {
      outfile = read_keyword(dl2monrc,DL2_KEY_MONOUT);
      if (outfile == NULL) {
        message("error: reading mon-output from %s\n",dl2monrc);
        return -1;
      } options |= SEL_OUTFCREAT;
    }
    if ((options & SEL_NETIFACE) == 0) {
      ifacename = read_keyword(dl2monrc,DL2_KEY_INTERFACE);
      if (ifacename == NSTR) ifacename = DL2_KEY_INTERFACE_eth0;
    }
  }

  // check socket-type:
  if (options & SEL_SOCKTYPE) {
    if (strcmp(socket_type_name,DL2_KEY_SOCKTYPE_RAW) == 0) {
      socket_type = DL2_ST_RAW;
    }
    else if (strcmp(socket_type_name,DL2_KEY_SOCKTYPE_DGRAM) == 0) {
      socket_type = DL2_ST_DGRAM;
    }
    else {
      message("error: unknown socket-type %s\n",socket_type_name);
      return -1;
    }
  } //else: use default 

  // check protocol:
  if (options & SEL_PROTNAME) {
    if (strcmp(protocol_name,DL2_KEY_PROTOCOL_ALL) == 0) {
      protocol = DL2_PD_ALL;
    }
    else if (strcmp(protocol_name,DL2_KEY_PROTOCOL_PPPD) == 0) {
      protocol = DL2_PT_PPP_DISC;
    }
    else if (strcmp(protocol_name,DL2_KEY_PROTOCOL_PPPS) == 0) {
      protocol = DL2_PT_PPP_SES;
    }
    else {
      message("error: unknown protocol %s\n",protocol_name);
      return -1;
    }
  } //else: use default

  // check for offline, test outfile:
  offline = (options & SEL_OFFLINE);
  if (offline != 0) {
    if (options & SEL_OUTFILE) {
      int flags = O_CREAT|O_WRONLY;
      if ((options & SEL_OUTFCREAT) == 0) flags |= O_EXCL;
      monfd = open(outfile,flags,00644);
      if (monfd < 0) {
        if (errno == EEXIST) {
          message("error: file exists %s\n",outfile);
        }
        else message("error: opening file %s (%d)\n",outfile,errno);
        return -1;
      } 
    }
    else {
      message("error: missing outfile (offline)\n");
      return -1;
    }
  }

  // initialize shared memory:
  if (offline == 0) {
    if (attach_shm(DL2_SHMPATHNAME) < 0) {
      message("error: attaching shm %s\n",DL2_SHMPATHNAME);
      return -1;
    }
  }

  // create a signal handler for SIGTERM,..:
  signal(SIGTERM,on_terminate);
  signal(SIGINT,on_terminate);
  signal(SIGCHLD,on_terminate);  

  // online: create child and connect by pipe (p.stdout->c.stdin):
  if (offline == 0) {
    const char *dl2pa = "dl2pa";
    int fd[2], i = 0;

    if (pipe(fd) < 0) {
      message("error: creating pipe (%d)\n",errno);
      return -1;      
    }

    analyzer_argv[i++] = dl2pa; 
    if (options & SEL_ANALYZER) i += n_analyzer_opt;
    if (options & SEL_RCFILE) {
      analyzer_argv[i++] = OPT_RCFILE; analyzer_argv[i++] = dl2monrc;
    } analyzer_argv[i++] = NULL;

    if ((cpid = fork()) == 0) {

      // child reads from stdin:
      close(fd[1]); close(0); monfd = dup(fd[0]);
      if (execvp(dl2pa,analyzer_argv) < 0) {
        message("error: executing %s (%d)\n",dl2pa,errno);
        return -1; 
      }
    }
    if (cpid < 0) {
      message("error: forking child (%d)\n",errno);
      return -1; 
    }
    // parent writes to stdout:
    close(fd[0]); close(1); monfd = dup(fd[1]);

    // signal here-we-are and wait for response:
    shmmonpa->flags |= DL2_SHMF_monup;
    while (42) {
      if (terminate) goto main_terminate;
      if (shmmonpa->flags & DL2_SHMF_paup) break; else sleep(1);
    }
  }

  if (monfd > 0) {
    Dl2_Monheader header;

    // setup output header:
    memset(&header,0,sizeof(header));
    header.socket_type = socket_type;
    header.protocol = protocol;
    if (ifacename != NULL) {
      int n = sizeof(header.interface);
      strncpy(header.interface,ifacename,n-1);
    }
    write(monfd,&header,sizeof(header));

    // access packet socket:
    sockfd = (rc = dl2_open(socket_type,protocol,ifacename));
    if (sockfd >= 0) {
      message("dl2mon started\n");
      rc = dl2_receive(sockfd,monfd);
      dl2_close(sockfd); sockfd = -1;
    }

  main_terminate:
    on_terminate(-1);
    message("dl2mon termed (%d)\n",rc);
    return rc;
  } 
  else {
    if (shmmonpa != NULL) detach_shm();    
    return -1;
  }
} //main 


//-locals---------------------------------------------------------------------
/* search for .dl2monrc, write absolute pathname to s; return -1 on error;
 * search order: environment, HOME.
*/
const char* locate_dl2monrc(void) {
  const char *dl2mondir = getenv(ENVIRON);
  if (dl2mondir == NULL) {
    dl2mondir = getenv("HOME");
  }
  if (dl2mondir != NULL) {
    char *s = malloc(strlen(dl2mondir)+strlen(RCFILE)+2);
    sprintf(s,"%s/%s",dl2mondir,RCFILE);
    return s;
  } return NULL;  
} //locate_dl2monrc


/* search rcfile cf for keyword key;
 * return its value or NSTR, NULL on error;
*/
const char* read_keyword(const char *cf,const char *key) {
  FILE *f = fopen(cf,"r");
  char *value = NULL;
  if (f != NULL) {
    char line[512], *p, *q;
    while (fgets(line,512,f) != NULL) {
      p = str2literal(line,NULL);
      if (p[0] == COMMENT_MARK || p[0] == 0) continue;
      if ((q = strchr(p,PARAMST_MARK)) == NULL) continue;
      *q = 0;
      if (strcmp(p,key) == 0) {
        p = str2literal(q+1,&q);        
        if (q > p) {
          value = malloc(q-p+1);
          strncpy(value,p,q-p);  
        } else value = NSTR; 
      }
    }
    if (ferror(f) != 0) {
      if (value != NULL) free(value);
      value = NULL;
    }    
    fclose(f);
  } return value;
} //read_keyword


char* str2literal(char* p,char** q) {
  char *r = p + strspn(p," \t\n"); 
  if (q != NULL) *q = r + strcspn(r," \t\n"); 
  return r;
} //str2literal


//-locals---------------------------------------------------------------------
/* attach to the memory shared with the DL2 Analyzer: set shmid, shmmonpa;
 * the process creates shm and attaches in RW mode; on success, it sets the
 * shm-segment's gid to that one of the shm-pathname in order to allow both 
 * the peers to access fairly; it is considered an error if there is already 
 * a Monitor attached to the shm-segment; return -1 on error;
*/
int attach_shm(const char *pn) {
  int trials, rc;
  key_t key = ftok(pn,1);
  if (key < 0) return -1;
  for (trials = 0; trials < 1; trials++) {
    shmid = shmget(key,sizeof(Dl2_Monpashm),IPC_CREAT|DL2_SHMV_MODE);
    if (shmid >= 0) {
      struct stat shmpn;
      struct shmid_ds ds;
      if ((rc = stat(pn,&shmpn)) < 0) break;
      if ((rc = shmctl(shmid,IPC_STAT,&ds)) < 0) break;
      if (ds.shm_nattch > 1) break;
      ds.shm_perm.gid = shmpn.st_gid;
      ds.shm_perm.mode = DL2_SHMV_MODE; 
      if ((rc = shmctl(shmid,IPC_SET,&ds)) < 0) break;
      shmmonpa = (Dl2_Monpashm*)shmat(shmid,NULL,0);
      if ((int)shmmonpa == -1) {shmmonpa = NULL; break;}
      if (shmmonpa->flags & DL2_SHMF_monup) {
        shmdt(shmmonpa); break;
      }
      shmmonpa->monpid = getpid();
      return rc;
    } else sleep(5);
  } return -1;
} //attach_shm


/* detach from the shm-segment; if we are the final attached process,
 * then remove the shm-segment from system; always return 0;
*/
int detach_shm(void) {
  struct shmid_ds ds;
  int rc = shmctl(shmid,IPC_STAT,&ds);
  if (rc >= 0 && (ds.shm_perm.uid == geteuid())) {
    shmmonpa->flags &= ~DL2_SHMF_monup;
    shmdt(shmmonpa);
    if (ds.shm_nattch == 1) shmctl(shmid,IPC_RMID,NULL);
  }
  shmmonpa = NULL; shmid = -1;
  return 0;
} //detach_shm


int nattach_shm(int id) {
  struct shmid_ds ds;
  int rc = shmctl(id,IPC_STAT,&ds);
  return (rc >= 0 ? ds.shm_nattch : rc);
} //nattach_shm 


//-locals---------------------------------------------------------------------
void message(char *fmt,...) {
  char sfmt[256];
  va_list ap;
  va_start(ap,fmt);
  sprintf(sfmt,"**%s",fmt);
  vfprintf(stderr,sfmt,ap);
  va_end(ap);  
} //message


void print_usage(void) {
  printf("DL2 Monitor (monitor), Version %.2lf (%s)\n",
    DL2_VERSION_NO,DL2_VERSION_DATE);
  printf("usage: dl2mon {<monitor-option>} [%s {<analyzer-option>}]\n",OPT_ANALYZER);
  printf("monitor-option (default):\n\
%s <socket-type> to be used by packet socket (%s)\n\
%s <protocol-name> to be used by packet socket (%s)\n\
%s <protocol-number> for test, only (%s)\n\
%s <interface-name> to bind to packet socket (%s)\n\
%s <output-file> in offline mode (%s)\n\
%s <rcfile-name> for monitor and analyzer configurations (%s)\n\
%s{<flag>} (%s)\n\
",
OPT_SOCKTYPE,DL2_KEY_SOCKTYPE_raw,
OPT_PROTNAME,DL2_KEY_PROTOCOL_all,
OPT_PROTNUM,"",
OPT_NETIFACE,"from rcfile",
OPT_OUTFILE,"from rcfile",
OPT_RCFILE,DL2_RCFILE,
OPT_FLAGS,"");
} //print_usage


// invoke this procedure only once!
void on_terminate(int sig) {
  terminate = 1;
  if (cpid > 0) {kill(cpid,SIGTERM); wait(NULL);}
  if (monfd >= 0) close(monfd);
  if (sockfd >= 0) dl2_close(sockfd);
  if (shmmonpa != NULL) detach_shm();
  if (sig != -1) {
    fprintf(stderr,"**dl2mon signal-termed (%d)\n",sig);
    exit(0);
  }
} //on_terminate

/***EOM***/
